use crate::backend::types::{City, Road, ShortestPath};
use anyhow::anyhow;
use neo4rs::{query, Graph, Node, Path};


pub async fn shortest_path(
    graph: &Graph,
    city1: String,
    city2: String,
) -> anyhow::Result<ShortestPath> {
    if city1 == city2 {
        return Err(anyhow!("Two cities must be different"));
    }

    let mut result = graph.execute(
        query("MATCH (c1:City {name: $c1}), (c2:City {name: $c2})
WITH c1, c2
MATCH path = (c1)-[:ROAD*]->(c2)
WITH path, reduce(distance = 0, rel in relationships(path) | distance + rel.distance) AS totalDistance
ORDER BY totalDistance ASC
LIMIT 1
RETURN path, totalDistance AS distance")
            .param("c1", city1)
            .param("c2", city2)
    ).await?;

    return if let Ok(Some(row)) = result.next().await {
        let path: Path = row.get("path")?;
        let distance: i32 = row.get("distance")?;

        let city_names: Vec<String> = path
            .nodes()
            .iter()
            .map(|node| node.get::<String>("name").unwrap())
            .collect();
        Ok(ShortestPath {
            path: city_names,
            distance,
        })
    } else {
        Err(anyhow!("couldn't find a valid path"))
    };
}

pub async fn all_cities(graph: &Graph, limit: Option<i32>) -> anyhow::Result<Vec<String>> {
    let mut cities = vec![];

    let query = match limit {
        None => query("MATCH (c:City) RETURN c"),
        Some(limit) => query("MATCH (c:City) RETURN c LIMIT $limit").param("limit", limit)
    };

    let mut result = graph
        .execute(query)
        .await?;
    while let Ok(Some(row)) = result.next().await {
        let node: Node = row.get("c")?;
        let name: String = node.get("name")?;
        cities.push(name);
    }
    Ok(cities)
}

pub async fn create_road(graph: &Graph, msg: Road) -> anyhow::Result<()> {
    let mut txn = graph.start_txn().await?;
    txn.run_queries([
        query("MATCH (c1:City{name: $city1}) WITH c1 MATCH (c2:City{name: $city2}) WITH c1,c2 CREATE (c1)-[:ROAD{distance: $distance}]->(c2)")
            .param("city1", msg.city1name.clone())
            .param("city2", msg.city2name.clone())
            .param("distance", msg.distance.clone()),
    ]).await?;
    if !msg.oneway {
        txn.run_queries([
            query("MATCH (c1:City{name: $city1}) WITH c1 MATCH (c2:City{name: $city2}) WITH c1,c2 CREATE (c1)<-[:ROAD{distance: $distance}]-(c2)")
                .param("city1", msg.city1name.clone())
                .param("city2", msg.city2name.clone())
                .param("distance", msg.distance.clone())
        ]).await?;
    }
    txn.commit().await?;
    Ok(())
}

pub async fn create_city(graph: &Graph, msg: City) -> anyhow::Result<()> {
    let mut txn = graph.start_txn().await?;
    txn.run_queries([query("CREATE (c:City {name: $name})").param("name", msg.name.clone())])
        .await?;
    txn.commit().await?;
    Ok(())
}
